/*
File ucd2utf8.c

Utility to generate from alist of Unicode hex values a C-language
definition with an array of pointers to UTF-8 strings.


Author and copyright 2020 Dr. Thomas redelberger, redetho@gmx.de
Developed with MinGW under Windows 10
True tabs meant to represent 4 spaces
*/

/* $Id$ */


#include <stdint.h>
#include <stdio.h>

/**
 * Encode a code point using UTF-8
 * 
 * @author Ondřej Hruška <ondra@ondrovo.com>
 * @license MIT
 * 
 * @param out - output buffer (min 5 characters), will be 0-terminated
 * @param utf - code point 0-0x10FFFF
 * @return number of bytes on success, 0 on failure (also produces U+FFFD, which uses 3 bytes)
 */
int utf8_encode(unsigned char *out, uint32_t utf) {
	if        (utf <= 0x7F    ) { /* Plain ASCII */
		out[0] = (unsigned char) utf;
		out[1] = 0;
		return 1;
	} else if (utf <= 0x07FF  ) { /* 2-byte unicode */
		out[0] = (unsigned char) (((utf >> 6) & 0x1F) | 0xC0);
		out[1] = (unsigned char) (((utf >> 0) & 0x3F) | 0x80);
		out[2] = 0;
		return 2;
	} else if (utf <= 0xFFFF  ) { /* 3-byte unicode */
		out[0] = (unsigned char) (((utf >> 12) & 0x0F) | 0xE0);
		out[1] = (unsigned char) (((utf >>  6) & 0x3F) | 0x80);
		out[2] = (unsigned char) (((utf >>  0) & 0x3F) | 0x80);
		out[3] = 0;
		return 3;
	} else if (utf <= 0x10FFFF) { /* 4-byte unicode */
		out[0] = (unsigned char) (((utf >> 18) & 0x07) | 0xF0);
		out[1] = (unsigned char) (((utf >> 12) & 0x3F) | 0x80);
		out[2] = (unsigned char) (((utf >>  6) & 0x3F) | 0x80);
		out[3] = (unsigned char) (((utf >>  0) & 0x3F) | 0x80);
		out[4] = 0;
		return 4;
	} else { /* error - use replacement character */
		out[0] = (unsigned char) 0xEF;
		out[1] = (unsigned char) 0xBF;
		out[2] = (unsigned char) 0xBD;
		out[3] = 0;
		return 0;
	}
}


#define BUFLEN	100		/* maximum line length */

int main(void) {
	char			buffer[BUFLEN+1];
	size_t			len;
	unsigned long	tmpULng;
	uint32_t 		unicode;
	unsigned char 	utf8_buf[5];
	int				utf8_len;
	int 			i;

	buffer[BUFLEN] = '\0';		/* sentinel */

	fputs("const char *unicode2utf8[128] = {\n", stdout);
	

	while (fgets(buffer, BUFLEN, stdin) != NULL) {	/* for each line */
		len = strlen(buffer);	/* fgets is not optimal but standard */

		if (buffer[len-1] == '\n') 	/* \n indicates that the line did fit */
			buffer[--len] = '\0';	/* but \n is not useful => kill it */

		tmpULng = strtoul(buffer, NULL, 16);	/* AtariST: 32bit, elsewhere: 64bit */
		unicode = (uint32_t) tmpULng;
		utf8_len = utf8_encode(utf8_buf, unicode);
  
		fputs("\t\"", stdout);
		for (i=0; i<utf8_len; i++)	/* output the utf-8 octets */
			printf("\\x%x", utf8_buf[i]);
		printf("\", /* %s */\n", buffer);	/* put orginal string as a C comment */
	}	/* while more lines */

	
	fputs("};\n", stdout);
	return 0;
}
